#include "controller.h"
#include "esmp3.h"

void Controller::handle() {
	if (last_rfid_check + 500 < millis() || last_rfid_check > millis()) {
		handle_rfid();
		last_rfid_check = millis();
	}
	handle_buttons();
}

void Controller::handle_buttons() {
	if (is_button_pressed(PIN_BTN_VOL_UP)) {
		log_i("BTN_VOL_UP pressed");
		uint8_t vol = min(audio.getVolume()+1, 21);
		log_d("Setting new volume %d", vol);
		audio.setVolume(vol);
	} else if (is_button_pressed(PIN_BTN_VOL_DOWN)) {
		log_i("BTN_VOL_DOWN pressed");
		uint8_t vol = max(audio.getVolume()-1, 1);
		log_d("Setting new volume %d", vol);
		audio.setVolume(vol);
	} else if (is_button_pressed(PIN_BTN_TRACK_NEXT)) {
		log_i("BTN_TRACK_NEXT pressed");
		next_track();
	} else if (is_button_pressed(PIN_BTN_TRACK_PREV)) {
		log_i("BTN_TRACK_PREV pressed");
		prev_track();
	}
}

void Controller::handle_rfid() {
	if (is_rfid_present) {
		byte buffer[2];
		byte buffer_size = 2;
		MFRC522Constants::StatusCode status = rfid->PICC_WakeupA(buffer, &buffer_size);
		if (status == MFRC522Constants::STATUS_OK) {
			// Card is still present.
			rfid->PICC_HaltA();
		} else {
			Serial.printf("RFID status is %s\n", MFRC522Debug::GetStatusCodeName(status));
			is_rfid_present = false;
			Serial.println("No more RFID card.\n");
			stop();
		}
	} else {
		if (rfid->PICC_IsNewCardPresent()) {
			if (rfid->PICC_ReadCardSerial()) {
				uint32_t uid = rfid->uid.uidByte[0]<<24 | rfid->uid.uidByte[1]<<16 | rfid->uid.uidByte[2]<<8 | rfid->uid.uidByte[3];
				Serial.printf("Found new rfid card with uid %x\n", uid);
				is_rfid_present = true;
				if (uid > 0) {
					String temp = String(uid, HEX);
					String s_uid = "";
					for (int i=0; i<(8-temp.length()); i++) {
						s_uid.concat("0");
					}
					s_uid.concat(temp);

					String data = read_rfid_data();

					play(s_uid);
				}
				rfid->PICC_HaltA();
			}
		}
	}
}

String Controller::read_rfid_data() {
	log_v("read_rfid_data() running...");
	static MFRC522::MIFARE_Key keys[8] = {
		{{0xd3, 0xf7, 0xd3, 0xf7, 0xd3, 0xf7}}, // D3 F7 D3 F7 D3 F7
		{{0xff, 0xff, 0xff, 0xff, 0xff, 0xff}}, // FF FF FF FF FF FF = factory default
		{{0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5}}, // A0 A1 A2 A3 A4 A5
		{{0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5}}, // B0 B1 B2 B3 B4 B5
		{{0x4d, 0x3a, 0x99, 0xc3, 0x51, 0xdd}}, // 4D 3A 99 C3 51 DD
		{{0x1a, 0x98, 0x2c, 0x7e, 0x45, 0x9a}}, // 1A 98 2C 7E 45 9A
		{{0xaa, 0xbb, 0xcc, 0xdd, 0xee, 0xff}}, // AA BB CC DD EE FF
		{{0x00, 0x00, 0x00, 0x00, 0x00, 0x00}}  // 00 00 00 00 00 00
	};
	log_d("Trying to read RFID data...");
	
	String data = "";
	MFRC522::PICC_Type type = rfid->PICC_GetType(rfid->uid.sak);
	
	uint8_t sectors = 0;
	switch(type) {
		case MFRC522Constants::PICC_TYPE_MIFARE_MINI: sectors = 5; break;
		case MFRC522Constants::PICC_TYPE_MIFARE_1K: sectors = 16; break;
		case MFRC522Constants::PICC_TYPE_MIFARE_4K: sectors = 40; break;
		default: log_i("Unknown PICC type %s\n", String(MFRC522Debug::PICC_GetTypeName(type)).c_str());
	}
	sectors = 2; // Pretend we have only two sectors, so we read only sector #1.
	for (uint8_t sector=1; sector<sectors; sector++) {
		uint8_t blocks = (sector < 32) ? 4 : 16;
		uint8_t block_offset = (sector < 32) ? sector * 4 : 128 + (sector - 32) * 16;
		
		MFRC522::StatusCode status;

		for (int i=0; i<8; i++) {
			MFRC522::MIFARE_Key *k = &keys[i];
			log_v("Trying MIFARE key %02X %02X %02X %02X %02X %02X...\n", k->keyByte[0], k->keyByte[1], k->keyByte[2], k->keyByte[3], k->keyByte[4], k->keyByte[5]);
			status = rfid->PCD_Authenticate(MFRC522Constants::PICC_CMD_MF_AUTH_KEY_A, block_offset, k, &rfid->uid);
			if (status == MFRC522Constants::STATUS_OK) {
				log_v("Authentication succeeded with key #%d\n", i);
				break;
			}
		}
		for (uint8_t block=0; block<blocks-1; block++) {
			byte buffer[18];
			uint8_t byte_count = 18;
			status = rfid->MIFARE_Read(block_offset + block, buffer, &byte_count);
			if (status != MFRC522Constants::STATUS_OK) {
				log_d("MIFARE_Read() failed: %s\n", String(MFRC522Debug::GetStatusCodeName(status)).c_str());
				continue;
			}
			for (int i=0; i<16; i++) {
				if (buffer[i]>=0x20 && buffer[i]<0x7F) data.concat((char)buffer[i]);
			}
		}
	}
	
	//_rfid->PICC_HaltA();
	rfid->PCD_StopCrypto1();
	log_d("Data from RFID: %s\n", data.c_str());
	return data;
}

void Controller::play(String rfid_id) {
	if (!rfid_id.equals(current_playlist.get_rfid_id())) {
		if (pm->has_playlist(rfid_id)) {
			current_playlist = pm->get_playlist(rfid_id);
			play();
		} else {
			Serial.printf("There is no playlist for rfid_id %s\n", rfid_id.c_str());
		}
	} else {
		if (!audio.isRunning()) {
			play();
		}
	}
}

void Controller::play() {
	String file = current_playlist.get_current_file_name();
	Serial.printf("Playing file: %s\n", file.c_str());
	audio.connecttoFS(SD, file.c_str(), current_playlist.get_current_time());
}

void Controller::next_track() {
	if (current_playlist.next_track()) {
		play();
	}
}

void Controller::prev_track() {
	uint32_t time = audio.getAudioCurrentTime();
	log_d("prev_track() called. getAudioCurrentTime() returns %d", time);
	if (time >= 5) {
		log_d("Restarting current track.");
		current_playlist.restart();
		play();
	} else {
		if (current_playlist.prev_track()) {
			play();
		}
	}
}

void Controller::stop() {
	if (audio.isRunning()) {
		current_playlist.set_current_time(audio.stopSong());
	}
}

bool Controller::is_button_pressed(uint8_t pin) {
	//log_d("Button %d state is %d", pin, digitalRead(pin));
	if (!digitalRead(pin)) {
		// Button is pressed - let's debounce it.
		if (button_pressed == pin) {
			if (button_pressed_since + 150 < millis() && !button_already_processed) {
				button_already_processed = true;
				return true;
			}
		} else {
			button_pressed = pin;
			button_pressed_since = millis();
			button_already_processed = false;
		}
	} else {
		if (button_pressed == pin) {
			button_pressed = 0;
		}
	}
	return false;
}